#!/usr/bin/env/python

"""
Doodle: Location Game
by Kris Schnee


"""

__author__ = "Kris Schnee"
__license__ = ""

##### DEPENDENCIES #####

## Standard
import math
import random

## Third-party
import pygame
##from pygame.locals import *

## Mine
import ringtale
import art
pygame.display.set_icon(pygame.image.load("acorn.png"))

##### CONSTANTS #####

CLOSE_DISTANCE = 10
COMBAT_DISTANCE = 15
TRAIL_COLOR_DEFAULT = (255,255,255)
art.fonts["small"] = art.fonts["EagerNaturalist14"]

WORLD_SIZE = (640,500)

##### CLASSES #####

class TrailDot:
    def __init__(self,coords,ttl,color=TRAIL_COLOR_DEFAULT):
        self.coords = (coords[0],coords[1]) ## Avoids direct reference
        self.ttl = ttl
        self.color = color




class Place:
    def __init__(self,**options):
        self.name = options.get("name","Place")
        self.coords = list(options.get("coords",[0,0]))
        self.big = options.get("big",False)

        ## Graphical options.
        self.nametag = None
        self.color = options.get("color",(255,255,255))




class Weapon:
    def __init__(self,**options):
        self.name = options.get("name","Weapon")
        self.weapon_type = options.get("weapon_type","sword")
        self.level = options.get("level",1)

    def Upgrade(self):
        self.level += 1

    def GetAttackStrength(self):
        return int(pow(1.2,self.level-1))




class Creature:
    def __init__(self,**options):
        self.name = options.get("name","Grignr")
        self.party = None

        ## Attacks get 20% more powerful per weapon level.
        self.weapon = Weapon(name=options.get("weapon_name","Sword"))

        self.HP = options.get("HP",100)
        self.HP_max = options.get("HP",100)
        self.XP = 0
        self.level = 1
        self.speed = options.get("speed",1)

        self.image = options.get("image","default")

    def GetAttackStrength(self):
        return 10 * self.level * self.weapon.GetAttackStrength()

    def GetXP(self,amount):
        self.XP += amount
        if self.XP >= 100:
            self.XP = 0
            self.level += 1
            self.HP_max += 10

            ## Upgrade graphics. Should only apply to PC!
            if self.level == 2:
                self.image = "centaur_sword"
                self.party.UpdateImage()
            elif self.level == 3:
                self.image = "centaur_sword_armor"
                self.party.UpdateImage()

    def TakeDamage(self,amount):
        """Returns whether HP > 0 after damage."""
        if amount > self.HP:
            self.HP = 0
            return False
        else:
            self.HP -= amount
            return True

    def Heal(self,amount=0):
        if amount:
            self.HP += amount
        else:
            self.HP = self.HP_max




class Party:
    def __init__(self,**options):
        self.members = []
        self.coords = list(options.get("coords",[0,0]))
        self.destination = None ## Eg. [0,0]
        self.moving = False
        self.in_town = False
        self.speed = options.get("speed",1)

        ## A trail of visible points left by the party.
        self.trail = [] ## Eg. [(4,2),(10,4),...]
        self.trail_counter = 0 ## A dot is left every 10 movement ticks.
        self.trail_color = options.get("trail_color",(255,32,32))
        self.trail_ttl = options.get("trail_ttl",8) ## Stealthy == low

        self.vision_range = 100
        self.known_places = []

        self.money = 0

        self.loot = 25
        self.loot_max = 100

        self.supplies = 75
        self.supplies_max = 100

        self.is_npc = options.get("is_npc",True) ## Computer-controlled?

        self.image = options.get("image","default") ## A sprite image name.

    def Add(self,who):
        self.members.append(who)
        who.party = self ## Give them a link back to me.
        self.speed = min([x.speed for x in self.members])
        self.UpdateImage()

    def SetKnownPlaces(self,place_list):
        """The party will have access to info on these places."""
        self.known_places = [item for item in place_list]

    def SetDestination(self,where):
        where = SanityCheckCoords(where)
        self.destination = where
        self.moving = True

    def TakeDamage(self,amount):
        """Apply damage to every party member.

        Individual TakeDamage function returns whether character is alive.
        This function doesn't use that return info."""
        
        for m in self.members:
            m.TakeDamage(amount)

    def Heal(self,amount=0):
        for m in self.members:
            m.Heal(amount)

    def LoseSupplies(self,amount):
        """Returns True if remaining supplies > 0.

        Supplies are a variable of the whole party."""
        
        if amount > self.supplies:
            self.supplies = 0
            self.TakeDamage(1)
            return False
        else:
            self.supplies -= amount
            return True

    def GetSupplies(self,amount=0):
        if amount:
            self.supplies = min(self.supplies_max, self.supplies+amount)
        else:
            self.supplies = self.supplies_max

    def GetXP(self,amount):
        if self.is_npc:
            return ## No XP for the wicked.
        for m in self.members:
            m.GetXP(amount)

    def GetLoot(self,amount,allow_spillover=False):
        """Adds this amount iff there's room. Returns True if OK.

        If allow_spillover, loot always is added, capped at loot_max."""
        
        if self.loot + amount > self.loot_max:
            if allow_spillover:
                self.loot = self.loot_max
                return True
            else:
                return False
        else:
            self.loot += amount
            return True

    def GetMoney(self,amount):
        self.money += amount

    def SpendMoney(self,amount):
        """Reduce money iff amount < self.money; returns True if spent."""
        if self.amount > self.money:
            return False
        else:
            self.money -= amount
            return True

    def MarkTrail(self):
        self.trail_counter += 1
        if self.trail_counter == 10:
            self.trail_counter = 0
            self.trail.append(
                TrailDot(self.coords,
                         self.trail_ttl,
                         self.trail_color)
                )
            ## Delete old dots.
            for dot in self.trail:
                dot.ttl -= 1
            self.trail = [dot for dot in self.trail if dot.ttl]

    def UpdateImage(self):
        self.image = self.members[0].image



class NPCParty(Party):
    def __init__(self,**options):
        Party.__init__(self,**options)

        self.wait_timer = 0

    def AI(self):
        ## Cheat:
        self.GetSupplies() ## Infinite supplies
        
        self.wait_timer += 1
        if self.wait_timer == 10:
            self.wait_timer = 0

            if not self.destination:
                if random.randint(0,99) < 50:
                    ## Pick a random spot.
                    x = self.coords[0] + random.randint(-100,100)
                    y = self.coords[1] + random.randint(-100,100)
                    print self.members[0].name+": Setting destination: "+str((x,y))
                    self.SetDestination((x,y))
                else:
                    ## Pick a town or something to visit.
                    n = random.randint(0,len(self.known_places)-1)
                    place = self.known_places[n]
                    print self.members[0].name+": Setting destination: "+place.name
                    self.SetDestination(place.coords)




class World:
    def __init__(self):

        self.places = []
        self.place_index = {}

        self.parties = []
        self.player_party = None

    def MakePlace(self,**options):
        p = Place(**options)
        if p.name in self.place_index.keys():
            raise "Error: There's already a place by that name."
        self.places.append(p)
        self.place_index[p.name] = p

        ## Graphical stuff.
        p.nametag = art.Write(p.name)

    def MakeSampleWorld(self):
        self.MakePlace(name="Dreams-Of-Blue",coords=(140,25))
        self.MakePlace(name="Fort Transfixion",coords=(425,410))
        self.MakePlace(name="Tail Town",coords=(400,200))
        self.MakePlace(name="Griffin Keep",coords=(125,400),big=True)
        self.MakePlace(name="Great Oak",coords=(425,85),big=True)
        self.MakePlace(name="Ironleaf",coords=(525,85))

        self.player_party = Party(coords=(500,50),is_npc=False)
        self.player = Creature(name="Kay",image="centaur",speed=2)
        self.player_party.Add(self.player)

        self.parties.append(self.player_party)

        ep = NPCParty(coords=(200,100))
        ep.Add(Creature(name="Orcus",HP=75))
        self.parties.append(ep)
        ep.SetKnownPlaces(self.places)

        ep2 = NPCParty(coords=(500,350))
        ep2.Add(Creature(name="Wistrik the Evil",HP=75))
        self.parties.append(ep2)
        ep2.SetKnownPlaces(self.places)

        self.SpawnCentaurs()

    def CheckForCombat(self):
        """Minimal demo of a battle system."""
        if self.player_party.in_town:
            return ## No combat allowed while in town.
        any_party_slain = False
        c = self.player_party.coords
        for enemy_party in self.parties[1:]:
            if enemy_party.in_town:
                continue ## Never mind.
            if Distance(enemy_party.coords,c) < COMBAT_DISTANCE:
                ## Extremely abstracted (yet action-packed) fighting!
                attack = len(enemy_party.members) * 2
                defense = self.player_party.members[0].level
                self.player_party.TakeDamage(max(1,attack-defense)) ## >=1
                
                attack = self.player_party.members[0].level * 4
                defense = len(enemy_party.members)
                enemy_party.TakeDamage(attack-defense)
        
                if not enemy_party.members[0].HP:
                    any_party_slain = True
                    self.player_party.GetLoot(enemy_party.loot)
                    self.player_party.GetXP(len(enemy_party.members) * 25)

        ## Destroy defeated enemy parties.
        if any_party_slain:
            self.parties = [self.player_party] + [p for p in self.parties[1:] if p.members[0].HP]
                    

    def RandomlySpawnEnemies(self):
        how_many_enemies = len(self.parties) - 1
        n = random.randint(0,999)
        if (how_many_enemies == 0) or ((how_many_enemies == 1) and (n<50)) or ((how_many_enemies < 5) and (n<10)):
            self.SpawnSomeOrcs()

    def SpawnSomeOrcs(self):
        """Create a sample enemy party."""
        ## Pick a spot away from the player. (One will always exist.)
        x, y = self.player_party.coords
        while Distance((x,y),self.player_party.coords) < 200:
            x = random.randint(10,WORLD_SIZE[0]-10)
            y = random.randint(10,WORLD_SIZE[1]-10)
        ep = NPCParty(coords=(x,y))
        ep.Add(Creature(name="Orc"))
        
        ## Automatic level-scaling, just like Bethesda! =)
        possible_extra_orcs = min(self.player_party.members[0].level-1,3)
        if possible_extra_orcs:
            for n in range(random.randint(0,possible_extra_orcs)):
                ep.Add(Creature(name="Orc"))
        if len(ep.members) > 2:
            ## Switch to a more menacing graphic.
            ep.members[0].image = "orc_sword"
            ep.UpdateImage()
        self.parties.append(ep)

        ep.GetLoot(len(ep.members)*10)
        ep.SetKnownPlaces(self.places)        

    def SpawnCentaurs(self):
        """Spawn a higher-level enemy party."""
        ## Pick a spot away from the player. (One will always exist.)
        x, y = self.player_party.coords
        while Distance((x,y),self.player_party.coords) < 200:
            x = random.randint(10,WORLD_SIZE[0]-10)
            y = random.randint(10,WORLD_SIZE[1]-10)
        ep = NPCParty(coords=(x,y))
        ep.Add(Creature(name="Centaur Bandit",HP=50,speed=2,image="centaur_bandit"))
        ep.Add(Creature(name="Centaur Bandit",HP=50,speed=2,image="centaur_bandit"))
        ep.loot = 50
        self.parties.append(ep)
        ep.SetKnownPlaces([p for p in self.places if p.name in ("Griffin Keep","Fort Transfixion")])

    def RunWorld(self):
        for party in self.parties:

            if party.is_npc:
                party.AI()

            party.MarkTrail() ## Even if stationary. (Makes old trail fade.)

            if party.destination and party.moving:
                ## Move party toward destination if any.
                if Distance(party.coords,party.destination) < CLOSE_DISTANCE:
                    party.moving = False
                    party.destination = None
                else:
                    self.MoveParty(party)
##                    party.GetXP(1)
                    self.HealNearTowns(party)

        self.RandomlySpawnEnemies()
        self.CheckForCombat()

    def MoveParty(self,party):
        speed = party.speed
        x, y = party.coords
        dest_x, dest_y = party.destination
        if x < dest_x:
            party.coords[0] = party.coords[0] + speed
        elif x > dest_x:
            party.coords[0] = party.coords[0] - speed
        if y < dest_y:
            party.coords[1] = party.coords[1] + speed
        elif y > dest_y:
            party.coords[1] = party.coords[1] - speed

        party.LoseSupplies(1)

    def PartyIsNearTown(self,party):
        for place in self.places:
            if Distance(party.coords,place.coords) < CLOSE_DISTANCE:
                return True
        return False

    def HealNearTowns(self,party):
        if self.PartyIsNearTown(party):
            party.in_town = True
            party.Heal()
            party.GetSupplies()
            party.GetMoney(party.loot)
            party.loot = 0
        else:
            party.in_town = False




class Game(World,ringtale.Framework):
    def __init__(self):
        World.__init__(self)
        ringtale.Framework.__init__(self)

        pygame.display.set_caption("Planar RPG Demo")
        pygame.display.set_icon(art.posters["acorn"])

        self.screen = pygame.display.set_mode((800,500))
        
        self.gfx_destination_x = art.posters["gold_x_24"]
        self.gfx_destination_x.set_colorkey((0,0,0))

        self.clock = pygame.time.Clock()
        
        ## The world coordinates which are drawn as pixel (0,0):
        self.camera_x = 0
        self.camera_y = 0
        
        self.done = False

        self.MakeSampleWorld()

        self.PushState("Explore")

    def HandleEvents(self,events=[]):
        for event in events:
            if event.type == pygame.locals.QUIT:
                self.done = True

            elif event.type == pygame.locals.KEYDOWN:
                if event.key == pygame.locals.K_ESCAPE:
                    self.PopState()
                    return ## We're done.

                elif event.key == pygame.locals.K_j:
                    pass ## Etc

            elif event.type == pygame.locals.MOUSEBUTTONDOWN:
                ## Find where in the world was clicked.
                world_coords = self.GetClickedCoords(event.pos)

                self.player_party.SetDestination(world_coords)

    def GetClickedCoords(self,clicked_coords):
        return (clicked_coords[0]+self.camera_x,
                clicked_coords[1]+self.camera_y)

    def DrawMovementDots(self):
        for party in self.parties:
            for dot in party.trail:
                if (not party.is_npc) or (Distance(dot.coords,self.player_party.coords) <= self.player_party.vision_range):
                    coords = self.GetScreenCoords(dot.coords)
                    color = party.trail_color
                    radius = 2
                    pygame.draw.circle(self.screen,color,coords,radius)

    def AddMovementDot(self):
        self.movement_dots.append((self.player_party.coords[0],self.player_party.coords[1]))

    def DrawXPMeter(self):
        x = 640
        y = 60
        length = self.player_party.members[0].XP
        color = (32,200,200)
        pygame.draw.rect(self.screen,color,(x+1,y+1,length,18),0)
        ## Frame
        color = (255,255,255)
        pygame.draw.rect(self.screen,color,(x,y,100,20),1)
        ## Text
        t = art.Write("Level: "+str(self.player_party.members[0].level),"small")
        self.screen.blit(t,(x+10,y+2))

    def DrawMoneyMeter(self):
        x = 640
        y = 80
        t = art.Write("Money: "+str(self.player_party.money),"small")
        self.screen.blit(t,(x+10,y+2))

    def DrawLifeMeter(self):
        x = 640
        y = 20
        person = self.player_party.members[0]
        length = int(100*(float(person.HP) / float(person.HP_max)))
        color = (255,32,32)
        pygame.draw.rect(self.screen,color,(x+1,y+1,length,18),0)
        ## Frame
        color = (255,255,255)
        pygame.draw.rect(self.screen,color,(x,y,100,20),1)
        ## Text
        t = art.Write("HP: "+str(person.HP)+"/"+str(person.HP_max),"small")
        self.screen.blit(t,(x+10,y+2))

    def DrawSupplyMeter(self):
        x = 640
        y = 0
        length = int(100*(float(self.player_party.supplies) / float(self.player_party.supplies_max)))
        color = (32,180,32)
        pygame.draw.rect(self.screen,color,(x+1,y+1,length,18),0)
        ## Frame
        color = (255,255,255)
        pygame.draw.rect(self.screen,color,(x,y,100,20),1)
        ## Text
        t = art.Write("Supplies: "+str(self.player_party.supplies)+"/"+str(self.player_party.supplies_max),"small")
        self.screen.blit(t,(x+10,y+2))

    def DrawLootMeter(self):
        x = 640
        y = 40
        
        length = int(100*(float(self.player_party.loot) / float(self.player_party.loot_max)))
        color = (180,180,0)
        pygame.draw.rect(self.screen,color,(x+1,y+1,length,18),0)
        ## Frame
        color = (255,255,255)
        pygame.draw.rect(self.screen,color,(x,y,100,20),1)
        ## Text
        t = art.Write("Loot: "+str(self.player_party.loot)+"/"+str(self.player_party.loot_max),"small")
        self.screen.blit(t,(x+10,y+2))

    def ExploreDraw(self):
        self.screen.fill((0,0,0))
        self.screen.blit(art.posters["map_mithrol"],(0,0))

        self.DrawSupplyMeter()
        self.DrawLifeMeter()
        self.DrawLootMeter()
        self.DrawXPMeter()
        self.DrawMoneyMeter()
        self.DrawMovementDots()

        for place in self.places:
            if place.big:
                radius = 10
            else:
                radius = 5
            coords = self.GetScreenCoords(place.coords)
            pygame.draw.circle(self.screen,place.color,coords,radius)
            self.screen.blit(place.nametag,(coords[0]-(place.nametag.get_width()/2),coords[1]+16))

        ## Draw destination of player's party if any.
        if self.player_party.destination:
            coords = self.GetScreenCoords(self.player_party.destination)
            radius = 3
            self.screen.blit(self.gfx_destination_x,(coords[0]-12,coords[1]-12))

        ## Draw the party's leader if it's the player's or is in their sight.
        for party in self.parties:
            if party.is_npc and (
                Distance(party.coords,self.player_party.coords)
                > self.player_party.vision_range
                ):
                continue ## Don't draw; isn't in sight.

            ## Do draw.
            color = (0,192,255)
            radius = 3
            coords = self.GetScreenCoords(party.coords)

            ## Attempt 1: Draw a circle for the party.
##            pygame.draw.circle(self.screen,color,coords,radius)
            ## Attempt 2: Draw a still image. Kludgy.
            coords = (coords[0] - 12, coords[1] - 16)
            self.screen.blit(
                art.sprite_sheets[party.image]["image"],
                coords,
                art.sprite_sheets[party.image]["animations"]["stand"]["rects"][1][1][0]
                )
        
        pygame.display.update()

    def GetScreenCoords(self,world_coords):
        return (world_coords[0] - self.camera_x,
                world_coords[1] - self.camera_y)

    def ExploreLogic(self,events):
        self.HandleEvents(events)
        self.RunWorld()
        if self.player_party.members[0].HP == 0:
            self.PopState() ## End of game.


##### OTHER FUNCTIONS #####

def OrthoDistance(p1,p2):
    return abs(p1[0]-p2[0]) + abs(p1[1]-p2[1])

def Distance(p1,p2):
    return int(math.sqrt( pow(abs(p1[0]-p2[0]),2) + pow(abs(p1[1]-p2[1]),2) ))

def SanityCheckCoords(coords):
    x = min(WORLD_SIZE[0],max(0,coords[0]))
    y = min(WORLD_SIZE[1],max(0,coords[1]))
    return (x,y)

##### AUTORUN #####

if __name__ == "__main__": ## Do the following if this code is run by itself:
    pygame.init()
    g = Game()
    g.MainLoop()
    pygame.image.save(g.screen,"screenshot.bmp")
##    pygame.quit()
