#!/usr/bin/python
"""
Art
by Kris Schnee.
Loads all of a progam's graphics at once, using Pygame.
Assumes that the program has a subdirectory called "graphics", containing
several subdirs called "fonts", "tiles" &c. -- though these are optional.
The loaded files can be referenced as, eg., "art.tiles['brick']".
"""

##___________________________________
##          ~Header~
##___________________________________

__author__ = "Kris Schnee"
__version__ = "2011.11.28"
__license__ = "Public Domain"


##___________________________________
##      ~Imported Modules~
##___________________________________

## Standard
import os

## Third-party
import pygame
from pygame.locals import *

## Mine

##___________________________________
##            ~Constants~
##___________________________________

READABLE_GRAPHICS_FORMATS = ["bmp","png","gif","jpg"]
GRAPHICS_DIRECTORY = "graphics"
TILE_DIRECTORY = os.path.join(GRAPHICS_DIRECTORY,"tiles")
PORTRAIT_DIRECTORY = os.path.join(GRAPHICS_DIRECTORY,"portraits")
FONT_DIRECTORY = os.path.join(GRAPHICS_DIRECTORY,"fonts")
POSTER_DIRECTORY = os.path.join(GRAPHICS_DIRECTORY,"posters")
SPRITE_DIRECTORY = os.path.join(GRAPHICS_DIRECTORY,"sprites")
INTERFACE_DIRECTORY = os.path.join(GRAPHICS_DIRECTORY,"interface")

DEFAULT_FONT_SIZE = 32
TILE_SIZE = 64

COLORKEY = (255,0,255,255)

##___________________________________
##            ~Functions~
##___________________________________

def Inventory():
    text = ""
    """Show all available files."""
    text += "Portraits:"
    for key in portraits.keys():
        text += "\n  -"+key
    text += "\nTiles:"
    for key in tiles.keys():
        text += "\n  -"+key
    text += "\nPosters:"
    for key in posters.keys():
        text += "\n  -"+key
    text += "\nFonts:"
    for key in fonts.keys():
        text += "\n  -"+key
    text += "\nSprite sheets:"
    for key in sprite_sheets.keys():
        text += "\n  -"+key
    text += "\nInterface themes:"
    for key in interface.keys():
        text += "\n  -"+key
    return text

def Write(text,font="default",color=(255,255,255)):
    """Return rendered text; shortcut for pygame.font.render."""
    try:
        return fonts[font].render(text,1,color)
    except:
        raise SystemError("Font '"+font+"' not found.")

##___________________________________
##            ~Autorun~
##___________________________________


## Always create a display surface so that art can load, if none exists yet.
try:
    screen = screen
except:
    screen = pygame.display.set_mode((1,1))

"""
The directories we'll look for -- each is optional -- are:
-posters: Backgrounds, misc. illustrations
-portraits: Character faces
-icons: Mini portraits
-tiles: small images used to draw a landscape
-sprites: objects with animation frames and drawing functions
-fonts: Text designs.
-interface: Interface themes, for use with my widget system Driftwood.
"""

"""
    Fonts:
    If there's a text file matching the font file's name, load it, and use
    its contents to specify a set of sizes. Eg. if the file's contents are
    "42,36,12", three copies of the font will be made, stored as
    FontName42, FontName36, and FontName12. If there's no such file,
    only one copy will be made, with a default font size of DEFAULT_FONT_SIZE.
    Finally, if there's a text file called "default.txt", that will be used
    to specify a default font for the Write function.
"""
fonts = {}
if os.path.exists(FONT_DIRECTORY):
    pygame.font.init()
    ## Load each font in whatever sizes are listed in a matching text file.
    font_filenames = [f for f in os.listdir(FONT_DIRECTORY) if f.endswith("ttf")]
    font_textfiles = [f for f in os.listdir(FONT_DIRECTORY) if f.endswith("txt")]
    for font_name in font_filenames:
        if font_name[:-3]+"txt" in font_textfiles: ## If there's a matching txt,
            f = open(os.path.join(FONT_DIRECTORY,font_name[:-3]+"txt"),"r")
            sizes = f.read().replace(", ",",").split(",")
            f.close()
            for size in sizes:
                fonts[font_name[:-4]+size] = pygame.font.Font(os.path.join(FONT_DIRECTORY,font_name),int(size))
        else:
            print "Default size"
            fonts[font_name[:-4]] = pygame.font.Font(os.path.join(FONT_DIRECTORY,font_name),DEFAULT_FONT_SIZE)

    ## Which font should be the default, if any? Eg. "myfont32" = myfont @ 32
    ## Note: This should be one of the sizes explicitly loaded above.
    if "default.txt" in font_textfiles:
        f = open(os.path.join(FONT_DIRECTORY,"default.txt"),"r")
        default_font_name = f.read()
        if default_font_name.endswith(".ttf"):
            default_font_name = default_font_name[:-3]
        f.close()
        fonts["default"] = fonts[default_font_name]
else:
    print "no dir"


"""
    Portraits / Icons:
    Makes miniature versions of the portraits (tile-sized) while loading them.
"""
portraits = {}
icons = {}
if os.path.exists(PORTRAIT_DIRECTORY):
    image_filenames = [f for f in os.listdir(PORTRAIT_DIRECTORY) if f[-3:] in READABLE_GRAPHICS_FORMATS]
    for image_name in image_filenames:
        surface = pygame.image.load(os.path.join(PORTRAIT_DIRECTORY,image_name)).convert()
        surface.set_colorkey(COLORKEY)
        portraits[image_name[:-4]] = surface
        icons[image_name[:-4]] = pygame.transform.scale(surface,(TILE_SIZE,TILE_SIZE))

"""Posters"""
posters = {}
if os.path.exists(POSTER_DIRECTORY):
    image_filenames = [f for f in os.listdir(POSTER_DIRECTORY) if f[-3:] in READABLE_GRAPHICS_FORMATS]
    for image_name in image_filenames:
        surface = pygame.image.load(os.path.join(POSTER_DIRECTORY,image_name)).convert_alpha()
        posters[image_name[:-4]] = surface

"""Tiles"""
tiles = {}
if os.path.exists(TILE_DIRECTORY):
    image_filenames = [f for f in os.listdir(TILE_DIRECTORY) if f[-3:] in READABLE_GRAPHICS_FORMATS]
    for image_name in image_filenames:
        surface = pygame.image.load(os.path.join(TILE_DIRECTORY,image_name))
        tiles[image_name[:-4]] = surface

"""
    Sprite Sheets
    These are image files that each contain all animation frames for an
    animated 2D character. We'll look for a matching text file that describes
    how large each frame is, how many rows/columns to look for, and whether
    there's any "padding" between frames. (Padding currently not implemented.)

    For each text file in the sprite directory called "sprite_data_*.txt",
    this code will look for a graphics file called * (with various extensions)
    and build an entry in sprite_sheets based on both. An example of the format
    for sprite data files follows. The real thing should not be indented.

    ## Animation Data
    ## This file also serves as an example of the format used by the art module.
    ## The file is divided into paragraphs separated by blank lines.
    ## Any line/paragraph starting with "##" is ignored.
    ## Paragraph 0 is below. It contains basic info in key/value pairs (spaces OK).
    ## frame_size is the size of each animation frame in pixels; assumes no padding (grids around the frames). colorkey is a color to make transparent.

    frame_size: 24, 32
    colorkey: 32, 156, 255

    ## A list of animations.
    ## line 0: Name.
    ## line 1: What happens at the end (normally "freeze" to stop).
    ## line 2: Starting frame coordinates (in frames), from upper left.
    ## line 3: How many frames for each angle, and how many angles? (Typically 4,4)
    ## sprite_sheets["some_name"]["animations"] will contain entries by name,
    ## each with some number of Rect objects describing where the frames are
    ## within sprite_sheets["some_name"]["image"].

    stand
    freeze
    0,1
    1,4

    walk
    freeze
    0,2
    4,4

    (End of example)

"""
if os.path.exists(SPRITE_DIRECTORY):
    ## Load sprite sheets, and establish animation data for them.
    ## This includes animation rects keyed by action and angle.
    sprite_sheets = {}
    sprite_data_file_names = [f for f in os.listdir(SPRITE_DIRECTORY) if f.startswith("sprite_data_")]
    for data_file_name in sprite_data_file_names:

        ## Create a dictionary for data about the sprite.
        sprite_sheet = {"animations":{}}

        ## Get information about the sprite (eg. its animation frame size).
        f = open(os.path.join(SPRITE_DIRECTORY,data_file_name))
        sprite_data = f.read()
        f.close()

        ## Find an image with a matching name.
        image_file_name_base = data_file_name.split("sprite_data_")[-1].split(".")[0]
        image_file_name = [f for f in os.listdir(SPRITE_DIRECTORY) if (f[:-4] == image_file_name_base) and f[-3:] in READABLE_GRAPHICS_FORMATS]
        ## Sanity check
        possible_files = len(image_file_name)
        if possible_files == 0:
            raise SystemError("Error: Image file not found for sprite data file "+data_file_name+".")
        elif possible_files > 1:
            raise SystemError("Error: There are too many possible sprite images for data file "+data_file_name+".")
        ## Load the matching image.
        image_file_name = image_file_name[0]
        image = pygame.image.load(os.path.join(SPRITE_DIRECTORY,image_file_name)).convert()
        sprite_sheet["image"] = image

        ## Analyze information about this sprite.
        paragraphs = [p for p in sprite_data.split("\n\n") if p and not p.startswith("#")]
        if not paragraphs:
            raise SystemError("Error: There's no data about this sprite. (Sprite data file is malformed.)")

        ## Paragraph 0 is basic information.
        lines = [line for line in paragraphs[0].split("\n") if line and not line[0] == "#"]
        for line in lines:
            key, value = line.split(":")
            key = key.lstrip().rstrip()
            value = value.lstrip().rstrip()
            if "," in value:
                value = tuple([int(n) for n in value.replace(", ",",").split(",")])
            sprite_sheet[key] = value
        if not "frame_size" in sprite_sheet:
            raise SystemError("Error: Sprite "+image_file_name_base+" doesn't specify a frame size.")

        ## Apply transparency color (default or a specified one) to image.
        colorkey = sprite_sheet.get("colorkey",COLORKEY)
        if len(colorkey) == 3:
            colorkey = list(colorkey)
            colorkey.append(255)
        sprite_sheet["image"].set_colorkey( colorkey )

        ## Paragraphs 1+ describe animations for the character.
        frame_size = sprite_sheet["frame_size"]
        for paragraph in paragraphs[1:]:
            lines = [line for line in paragraph.split("\n") if line and not line[0] == "#"]
            anim_data = {}
            anim_data["name"] = lines[0]
            anim_data["end"] = lines[1]
            anim_start_coords = tuple([int(n) for n in lines[2].replace(", ",",").split(",")])
            anim_frames, anim_angles = tuple([int(n) for n in lines[3].replace(", ",",").split(",")])

            ## Build a list of clipping rectangles containing the anim. frames.
            frame_rects = []
            ox, oy = anim_start_coords
            angle_increment = 360 / anim_angles
            anim_data["angle_increment"] = angle_increment
            angle = 0
            for y in range(anim_angles):
                this_angle_rects = []
                for x in range(anim_frames):
                    this_angle_rects.append( (ox+(x*frame_size[0]),oy+(y*frame_size[1]),frame_size[0],frame_size[1]) )
                frame_rects.append( (angle,this_angle_rects) )
                angle += angle_increment
            anim_data["rects"] = frame_rects

            ## Done with this animation. Store this data for the sprite.
            sprite_sheet["animations"][anim_data["name"]] = anim_data

        ## Done! Store this info.
        sprite_sheets[image_file_name_base] = sprite_sheet


"""
Interface:
Try to load a set of border graphics and other data from each directory
under /graphics/interface. The other data should be in a text file called
"theme_info.txt", but this file is optional.
"""
interface = {}
if os.path.exists(INTERFACE_DIRECTORY):
    themes = os.listdir(INTERFACE_DIRECTORY)
    for theme in themes:
        interface[theme] = {"border_images":{}}
        image_filenames = [f for f in os.listdir(os.path.join(INTERFACE_DIRECTORY,theme)) if f[-3:] in READABLE_GRAPHICS_FORMATS]
        for image_name in image_filenames:
            surface = pygame.image.load(os.path.join(INTERFACE_DIRECTORY,theme,image_name))
            surface.set_colorkey(COLORKEY)
            interface[theme]["border_images"][image_name[:-4]] = surface
        if os.path.exists(os.path.join(INTERFACE_DIRECTORY,theme,"theme_info.txt")):
            f = open(os.path.join(INTERFACE_DIRECTORY,theme,"theme_info.txt"))
            theme_info = f.read()
            f.close()
            lines = [line for line in theme_info.split("\n") if line and not line[0] == "#"]
            if lines:
                for line in lines:
                    key, value = line.split(":")
                    key = key.lstrip().rstrip()
                    value = value.lstrip().rstrip()
                    if "," in value:
                        value = tuple([int(n) for n in value.split(",")])
                    interface[theme][key] = value

## Graphics are now loaded and accessible at art.fonts, art.tiles &c.
pygame.display.quit()
